@@ -8,10 +8,17 @@ import os
 import random
 import re
 import time
-from xml.etree import ElementTree
+from xml.etree import ElementTree  # noqa: ICN001
 
 import requests
-from pkg_resources import DistributionNotFound, get_distribution, parse_version
+
+try:
+    from importlib.metadata import PackageNotFoundError
+    from importlib.metadata import version as _get_version
+except ImportError:
+    # python < 3.8
+    from importlib_metadata import PackageNotFoundError
+    from importlib_metadata import version as _get_version
 
 try:
     from kazoo.client import KazooClient, KazooState
@@ -65,12 +72,9 @@ __author__ = "Daniel Lindsley, Joseph Kocherhans, Jacob Kaplan-Moss, Thomas Ried
 __all__ = ["Solr"]
 
 try:
-    pkg_distribution = get_distribution(__name__)
-    __version__ = pkg_distribution.version
-    version_info = pkg_distribution.parsed_version
-except DistributionNotFound:
+    __version__ = _get_version(__name__)
+except PackageNotFoundError:
     __version__ = "0.0.dev0"
-    version_info = parse_version(__version__)
 
 
 def get_version():
@@ -83,6 +87,10 @@ DATETIME_REGEX = re.compile(
 # dict key used to add nested documents to a document
 NESTED_DOC_KEY = "_childDocuments_"
 
+VALID_XML_CHARS_REGEX = re.compile(
+    "[^\u0020-\uD7FF\u0009\u000A\u000D\uE000-\uFFFD\U00010000-\U0010FFFF]+"
+)
+
 
 class NullHandler(logging.Handler):
     def emit(self, record):
@@ -205,23 +213,6 @@ def safe_urlencode(params, doseq=0):
     return urlencode(new_params, doseq)
 
 
-def is_valid_xml_char_ordinal(i):
-    """
-    Defines whether char is valid to use in xml document
-
-    XML standard defines a valid char as::
-
-    Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]
-    """
-    # conditions ordered by presumed frequency
-    return (
-        0x20 <= i <= 0xD7FF
-        or i in (0x9, 0xA, 0xD)
-        or 0xE000 <= i <= 0xFFFD
-        or 0x10000 <= i <= 0x10FFFF
-    )
-
-
 def clean_xml_string(s):
     """
     Cleans string from invalid xml chars
@@ -230,7 +221,7 @@ def clean_xml_string(s):
 
     http://stackoverflow.com/questions/8733233/filtering-out-certain-bytes-in-python
     """
-    return "".join(c for c in s if is_valid_xml_char_ordinal(ord(c)))
+    return VALID_XML_CHARS_REGEX.sub("", s)
 
 
 class SolrError(Exception):
@@ -295,8 +286,9 @@ class Results(object):
         self.qtime = decoded.get("responseHeader", {}).get("QTime", None)
         self.grouped = decoded.get("grouped", {})
         self.nextCursorMark = decoded.get("nextCursorMark", None)
-        self._next_page_query = self.nextCursorMark is not None \
-            and next_page_query or None
+        self._next_page_query = (
+            self.nextCursorMark is not None and next_page_query or None
+        )
 
     def __len__(self):
         if self._next_page_query:
@@ -319,6 +311,9 @@ class Solr(object):
     Optionally accepts ``decoder`` for an alternate JSON decoder instance.
     Default is ``json.JSONDecoder()``.
 
+    Optionally accepts ``encoder`` for an alternate JSON Encoder instance.
+    Default is ``json.JSONEncoder()``.
+
     Optionally accepts ``timeout`` for wait seconds until giving up on a
     request. Default is ``60`` seconds.
 
@@ -341,6 +336,7 @@ class Solr(object):
         self,
         url,
         decoder=None,
+        encoder=None,
         timeout=60,
         results_cls=Results,
         search_handler="select",
@@ -348,12 +344,14 @@ class Solr(object):
         always_commit=False,
         auth=None,
         verify=True,
+        session=None,
     ):
         self.decoder = decoder or json.JSONDecoder()
+        self.encoder = encoder or json.JSONEncoder()
         self.url = url
         self.timeout = timeout
         self.log = self._get_log()
-        self.session = None
+        self.session = session
         self.results_cls = results_cls
         self.search_handler = search_handler
         self.use_qt_param = use_qt_param
@@ -470,8 +468,8 @@ class Solr(object):
         :param handler: defaults to self.search_handler (fallback to 'select')
         :return:
         """
-        # specify json encoding of results
-        params["wt"] = "json"
+        # Returns json docs unless otherwise specified
+        params.setdefault("wt", "json")
         custom_handler = handler or self.search_handler
         handler = "select"
         if custom_handler:
@@ -508,11 +506,13 @@ class Solr(object):
         clean_ctrl_chars=True,
         commit=None,
         softCommit=False,
+        commitWithin=None,
         waitFlush=None,
         waitSearcher=None,
         overwrite=None,
         handler="update",
         solrapi="XML",
+        min_rf=None,
     ):
         """
         Posts the given xml or json message to http://<self.url>/update and
@@ -539,10 +539,14 @@ class Solr(object):
         if commit is None:
             commit = self.always_commit
 
+        if min_rf:
+            query_vars.append("min_rf=%i" % min_rf)
         if commit:
             query_vars.append("commit=%s" % str(bool(commit)).lower())
         elif softCommit:
             query_vars.append("softCommit=%s" % str(bool(softCommit)).lower())
+        elif commitWithin is not None:
+            query_vars.append("commitWithin=%s" % str(int(commitWithin)))
 
         if waitFlush is not None:
             query_vars.append("waitFlush=%s" % str(bool(waitFlush)).lower())
@@ -724,7 +728,10 @@ class Solr(object):
             return value
 
         if isinstance(value, (list, tuple)):
-            value = value[0]
+            result = [self._to_python(v) for v in value]
+            if isinstance(value, tuple):
+                result = tuple(result)
+            return result
 
         if value == "true":
             return True
@@ -835,10 +842,12 @@ class Solr(object):
 
         cursorMark = params.get("cursorMark", None)
         if cursorMark != decoded.get("nextCursorMark", cursorMark):
+
             def next_page_query():
                 nextParams = params.copy()
                 nextParams["cursorMark"] = decoded["nextCursorMark"]
                 return self.search(search_handler=search_handler, **nextParams)
+
             return self.results_cls(decoded, next_page_query)
         else:
             return self.results_cls(decoded)
@@ -907,13 +916,62 @@ class Solr(object):
         )
         return res
 
-    def _build_doc(self, doc, boost=None, fieldUpdates=None):
+    def _build_docs(self, docs, boost=None, fieldUpdates=None):
+        # if no boost needed use json multidocument api
+        #   The JSON API skips the XML conversion and speedup load from 15 to 20 times.
+        #   CPU Usage is drastically lower.
+        if boost is None:
+            solrapi = "JSON"
+            message = docs
+            # single doc convert to array of docs
+            if isinstance(message, dict):
+                # convert dict to list
+                message = [message]
+                # json array of docs
+            if isinstance(message, list):
+                # convert to string
+                cleaned_message = [
+                    self._build_json_doc(doc, fieldUpdates=fieldUpdates)
+                    for doc in message
+                ]
+                m = self.encoder.encode(cleaned_message).encode("utf-8")
+            else:
+                raise ValueError("wrong message type")
+        else:
+            solrapi = "XML"
+            message = ElementTree.Element("add")
+
+            for doc in docs:
+                el = self._build_xml_doc(doc, boost=boost, fieldUpdates=fieldUpdates)
+                message.append(el)
+
+            # This returns a bytestring. Ugh.
+            m = ElementTree.tostring(message, encoding="utf-8")
+            # Convert back to Unicode please.
+            m = force_unicode(m)
+
+        return (solrapi, m, len(message))
+
+    def _build_json_doc(self, doc, fieldUpdates=None):
+        if fieldUpdates is None:
+            cleaned_doc = {k: v for k, v in doc.items() if not self._is_null_value(v)}
+        else:
+            # id must be added without a modifier
+            # if using field updates, all other fields should have a modifier
+            cleaned_doc = {
+                k: {fieldUpdates[k]: v} if k in fieldUpdates else v
+                for k, v in doc.items()
+            }
+
+        return cleaned_doc
+
+    def _build_xml_doc(self, doc, boost=None, fieldUpdates=None):
         doc_elem = ElementTree.Element("doc")
 
         for key, value in doc.items():
             if key == NESTED_DOC_KEY:
                 for child in value:
-                    doc_elem.append(self._build_doc(child, boost, fieldUpdates))
+                    doc_elem.append(self._build_xml_doc(child, boost, fieldUpdates))
                 continue
 
             if key == "boost":
@@ -927,18 +985,25 @@ class Solr(object):
             else:
                 values = (value,)
 
+            use_field_updates = fieldUpdates and key in fieldUpdates
+            if use_field_updates and not values:
+                values = ("",)
             for bit in values:
+                attrs = {"name": key}
+
                 if self._is_null_value(bit):
-                    continue
+                    if use_field_updates:
+                        bit = ""
+                        attrs["null"] = "true"
+                    else:
+                        continue
 
                 if key == "_doc":
-                    child = self._build_doc(bit, boost)
+                    child = self._build_xml_doc(bit, boost)
                     doc_elem.append(child)
                     continue
 
-                attrs = {"name": key}
-
-                if fieldUpdates and key in fieldUpdates:
+                if use_field_updates:
                     attrs["update"] = fieldUpdates[key]
 
                 if boost and key in boost:
@@ -963,6 +1028,7 @@ class Solr(object):
         waitSearcher=None,
         overwrite=None,
         handler="update",
+        min_rf=None,
     ):
         """
         Adds or updates documents.
@@ -986,6 +1052,8 @@ class Solr(object):
 
         Optionally accepts ``overwrite``. Default is ``None``.
 
+        Optionally accepts ``min_rf``. Default is ``None``.
+
         Usage::
 
             solr.add([
@@ -1001,53 +1069,28 @@ class Solr(object):
         """
         start_time = time.time()
         self.log.debug("Starting to build add request...")
-        solrapi = "XML"
-        # if no commands (no boost, no atomic updates) needed use json multidocument api
-        #   The JSON API skipts the XML conversion and speedup load from 15 to 20 times.
-        #   CPU Usage is drastically lower.
-        if boost is None and fieldUpdates is None:
-            solrapi = "JSON"
-            message = docs
-            # single doc convert to array of docs
-            if isinstance(message, dict):
-                # convert dict to list
-                message = [message]
-                # json array of docs
-            if isinstance(message, list):
-                # convert to string
-                m = json.dumps(message).encode("utf-8")
-            else:
-                raise ValueError("wrong message type")
-        else:
-            message = ElementTree.Element("add")
-
-            if commitWithin:
-                message.set("commitWithin", commitWithin)
-
-            for doc in docs:
-                el = self._build_doc(doc, boost=boost, fieldUpdates=fieldUpdates)
-                message.append(el)
-
-            # This returns a bytestring. Ugh.
-            m = ElementTree.tostring(message, encoding="utf-8")
-            # Convert back to Unicode please.
-            m = force_unicode(m)
-
+        solrapi, m, len_message = self._build_docs(
+            docs,
+            boost,
+            fieldUpdates,
+        )
         end_time = time.time()
         self.log.debug(
             "Built add request of %s docs in %0.2f seconds.",
-            len(message),
+            len_message,
             end_time - start_time,
         )
         return self._update(
             m,
             commit=commit,
             softCommit=softCommit,
+            commitWithin=commitWithin,
             waitFlush=waitFlush,
             waitSearcher=waitSearcher,
             overwrite=overwrite,
             handler=handler,
             solrapi=solrapi,
+            min_rf=min_rf,
         )
 
     def delete(
@@ -1093,11 +1136,18 @@ class Solr(object):
             else:
                 doc_id = list(filter(None, id))
             if doc_id:
-                m = "<delete>%s</delete>" % "".join("<id>%s</id>" % i for i in doc_id)
+                et = ElementTree.Element("delete")
+                for one_doc_id in doc_id:
+                    subelem = ElementTree.SubElement(et, "id")
+                    subelem.text = one_doc_id
+                m = ElementTree.tostring(et)
             else:
                 raise ValueError("The list of documents to delete was empty.")
         elif q is not None:
-            m = "<delete><query>%s</query></delete>" % q
+            et = ElementTree.Element("delete")
+            subelem = ElementTree.SubElement(et, "query")
+            subelem.text = q
+            m = ElementTree.tostring(et)
 
         return self._update(
             m,
@@ -1229,7 +1279,7 @@ class Solr(object):
             raise
 
         try:
-            data = json.loads(resp)
+            data = self.decoder.decode(resp)
         except ValueError:
             self.log.exception("Failed to load JSON response")
             raise
@@ -1302,7 +1352,7 @@ class SolrCoreAdmin(object):
         if params is None:
             params = {}
         if headers is None:
-            headers = {}
+            headers = {"Content-Type": "application/x-www-form-urlencoded"}
 
         resp = requests.get(url, data=safe_urlencode(params), headers=headers)
         return force_unicode(resp.content)
@@ -1427,6 +1477,7 @@ class SolrCloud(Solr):
         zookeeper,
         collection,
         decoder=None,
+        encoder=None,
         timeout=60,
         retry_count=5,
         retry_timeout=0.2,
@@ -1446,6 +1497,7 @@ class SolrCloud(Solr):
         super(SolrCloud, self).__init__(
             url,
             decoder=decoder,
+            encoder=encoder,
             timeout=timeout,
             auth=self.auth,
             verify=self.verify,
@@ -1454,7 +1506,7 @@ class SolrCloud(Solr):
         )
 
     def _send_request(self, method, path="", body=None, headers=None, files=None):
-        for retry_number in range(0, self.retry_count):
+        for retry_number in range(self.retry_count):
             try:
                 self.url = self.zookeeper.getRandomURL(self.collection)
                 return Solr._send_request(self, method, path, body, headers, files)
